@@ -180,7 +180,7 @@ typedef struct _NVHsChannelEvoRec {
     } nv3d;
 
     struct {
-        NvU32 handle;
+        NvU32 handle[1];
     } nv2d;
 
     /*
@@ -277,7 +277,7 @@ typedef struct _NVHsChannelEvoRec {
      */
     NVRgLine1CallbackPtr pRgIntrCallback;
 
-#if NVKMS_PROCFS_ENABLE
+#if NVKMS_HEADSURFACE_STATS
 
     /*
      * We track statistics differently for SwapGroup and non-SwapGroup
@@ -342,7 +342,7 @@ typedef struct _NVHsChannelEvoRec {
     } statistics;
 #else
 #define NVKMS_HEADSURFACE_STATS_MAX_SEMAPHORES 0
-#endif /* NVKMS_PROCFS_ENABLE */
+#endif /* NVKMS_HEADSURFACE_STATS */
 
     /*
      * We need one semaphore for the non-stall interrupt following rendering to
@@ -481,6 +481,7 @@ static inline void HsIncrementNextIndex(
 }
 
 static inline void HsChangeSurfaceFlipRefCount(
+    NVDevEvoPtr pDevEvo,
     NVSurfaceEvoPtr pSurfaceEvo,
     NvBool increase)
 {
@@ -488,7 +489,7 @@ static inline void HsChangeSurfaceFlipRefCount(
         if (increase) {
             nvEvoIncrementSurfaceRefCnts(pSurfaceEvo);
         } else {
-            nvEvoDecrementSurfaceRefCnts(pSurfaceEvo);
+            nvEvoDecrementSurfaceRefCnts(pDevEvo, pSurfaceEvo);
         }
     }
 }
